#include "he86.h"
#include <stdio.h>

void he86_run(struct HE86_REGS *r)
{
	int c0, ti0, ti1;
	char msg[40];
	for (;;) {
		if (r->step <= 0) break;
		if ((r->sysflags & HE86_SFLG_HALT) != 0) break;
		if ((r->sysflags & HE86_SFLG_NMI) != 0 && (r->sysflags & HE86_SFLG_NMIMSK) == 0) {
			he86_int(r, 0x02); /* NMI */
		}
		if ((r->sysflags & HE86_SFLG_IRQ) != 0 && (r->eflags & HE86_FLG_IF) != 0) {
			(r->irq)(r);
		}
		/* Õ`FbNȂǂȂ̂ŁAVv */
		r->sysflags &= ~(HE86_SFLG_USE16OPR | HE86_SFLG_USE32OPR | HE86_SFLG_USE16ADR | HE86_SFLG_USE32ADR);
		if ((r->sysflags & HE86_SFLG_USE16C) != 0) {
			r->sysflags |= HE86_SFLG_USE16OPR | HE86_SFLG_USE16ADR;
		}
		if ((r->sysflags & HE86_SFLG_USE32C) != 0) {
			r->sysflags |= HE86_SFLG_USE32OPR | HE86_SFLG_USE32ADR;
		}
		r->defseg = 127;
		for (;;) {	/* vtBNXp̃[v */
			c0 = he86_fetch_code1(r); /* 11bytëӖ */
			if (0x40 <= c0 && c0 <= 0x47) {	/* INC */
				if ((r->sysflags & HE86_SFLG_USE16OPR) != 0) {
					r->reg32[c0 & 7] = (r->reg32[c0 & 7] & ~0xffff) | ((r->reg32[c0 & 7] + 1) & 0xffff);
				}
				if ((r->sysflags & HE86_SFLG_USE32OPR) != 0) {
					r->reg32[c0 & 7]++;
				}
				break;
			}
			if (0x48 <= c0 && c0 <= 0x4f) {	/* DEC */
				if ((r->sysflags & HE86_SFLG_USE16OPR) != 0) {
					r->reg32[c0 & 7] = (r->reg32[c0 & 7] & ~0xffff) | ((r->reg32[c0 & 7] - 1) & 0xffff);
				}
				if ((r->sysflags & HE86_SFLG_USE32OPR) != 0) {
					r->reg32[c0 & 7]--;
				}
				break;
			}
			if (0x50 <= c0 && c0 <= 0x57) {	/* PUSH */
				if ((r->sysflags & HE86_SFLG_USE16OPR) != 0) {
					he86_push2(r, r->reg32[c0 & 7] & 0xffff);
				}
				if ((r->sysflags & HE86_SFLG_USE32OPR) != 0) {
					he86_push4(r, r->reg32[c0 & 7]);
				}
				break;
			}
			if (c0 == 0x5c) { (*(r->panic))(r, "he86_run: POP(eSP);"); } /* POP(eSP); */
			if (0x58 <= c0 && c0 <= 0x5f) {	/* POP */
				/* ESPPOPꍇ̂Ƃ͂悭lĂȂ */
				if ((r->sysflags & HE86_SFLG_USE16OPR) != 0) {
					r->reg32[c0 & 7] = (r->reg32[c0 & 7] & ~0xffff) | (he86_pop2(r) & 0xffff);
				}
				if ((r->sysflags & HE86_SFLG_USE32OPR) != 0) {
					r->reg32[c0 & 7] = he86_pop4(r);
				}
				break;
			}
			if (c0 == 0x66) {
				r->sysflags ^= (HE86_SFLG_USE16OPR | HE86_SFLG_USE32OPR);
				continue;
			}
			if (c0 == 0x67) {
				r->sysflags ^= (HE86_SFLG_USE16ADR | HE86_SFLG_USE32ADR);
				continue;
			}
			if (c0 == 0x8e) {	/* MOV */
				ti0 = he86_fetch_code1(r);
				ti1 = he86_read_regmem(r, ti0, 2);
				he86_loadsreg(r, (ti0 >> 3) & 7, ti1, 0);
				break;
			}
			if (0x90 <= c0 && c0 <= 0x97) {	/* XCHG */
				if ((r->sysflags & HE86_SFLG_USE16OPR) != 0) {
					ti0 = r->reg32[HE86_EAX] & 0xffff;
					ti1 = r->reg32[c0 & 0x07] & 0xffff;
					r->reg32[HE86_EAX] = (r->reg32[HE86_EAX] & ~0xffff) | ti1;
					r->reg32[c0 & 0x07] = (r->reg32[c0 & 0x07] & ~0xffff) | ti0;
					break;
				}
				if ((r->sysflags & HE86_SFLG_USE32OPR) != 0) {
					ti0 = r->reg32[HE86_EAX];
					r->reg32[HE86_EAX] = r->reg32[c0 & 0x07];
					r->reg32[c0 & 0x07] = ti0;
					break;
				}
			}
			if (0xb0 <= c0 && c0 <= 0xb3) {	/* MOV */
				ti0 = he86_fetch_code1(r);
				r->reg32[c0 & 3] = (r->reg32[c0 & 3] & ~0xff) | (ti0 & 0xff);
				break;
			}
			if (0xb4 <= c0 && c0 <= 0xb7) {	/* MOV */
				ti0 = he86_fetch_code1(r);
				r->reg32[c0 & 3] = (r->reg32[c0 & 3] & ~0x0000ff00) | (ti0 & 0xff) << 8;
				break;
			}
			if (0xb8 <= c0 && c0 <= 0xbf) {	/* MOV */
				r->reg32[c0 & 7] = he86_fetch_code2(r);
				break;
			}
			if (c0 == 0xcc) {
				he86_int(r, 0x03);
				break;
			}
			if (c0 == 0xcd) {
				ti0 = he86_fetch_code1(r);
				he86_int(r, ti0);
				break;
			}
			if (c0 == 0xf4) {	/* HLT */
				r->sysflags |= HE86_SFLG_HALT;
				break;
			}
			/*  */
			sprintf(msg, "he86_run: op=0x%02X", c0);
			(*(r->panic))(r, msg);
		}
		r->step--;
	}
	return;
}

int he86_ladr(struct HE86_REGS *r, int sreg_no, int ofs)
{
	return r->segbase[sreg_no] + ofs;
}


void he86_int(struct HE86_REGS *r, int n)
{
	(*(r->panic))(r, "he86_int: ");
}

int he86_fetch_code1(struct HE86_REGS *r)
{
	if ((r->sysflags & HE86_SFLG_USE16C) != 0) {
		int ip = r->eip & 0xffff, a;
		a = ip + r->segbase[HE86_CS];
		ip++;
		r->eip = (r->eip & ~0xffff) | (ip & 0xffff);
		if ((r->cr[0] & HE86_CR0_PG) != 0) {
			(*(r->panic))(r, "he86_fetch_code1: USE16C.paging");
		}
		return (*(r->readmem))(r, a, 1);
	}
	if ((r->sysflags & HE86_SFLG_USE32C) != 0) {
		(*(r->panic))(r, "he86_fetch_code1: USE32C");
	}
	return -1;
}

int he86_fetch_code2(struct HE86_REGS *r)
{
	int c0, c1;
	c0 = he86_fetch_code1(r);
	c1 = he86_fetch_code1(r) << 8;
	return c0 | c1;
	/* {͂ƏāAreadmemwidth=2ŌĂяo */
}

void he86_push2(struct HE86_REGS *r, int i)
{
	(*(r->panic))(r, "he86_push2: ");
}

void he86_push4(struct HE86_REGS *r, int i)
{
	(*(r->panic))(r, "he86_push4: ");
}

int he86_pop2(struct HE86_REGS *r)
{
	(*(r->panic))(r, "he86_pop2: ");
	return 0;
}

int he86_pop4(struct HE86_REGS *r)
{
	(*(r->panic))(r, "he86_pop4: ");
	return 0;
}

unsigned int he86_read_regmem(struct HE86_REGS *r, int modrm, int width)
{
	char mod = modrm >> 6;
	int v = 0;
	modrm &= 7;
	if (mod == 3) {	/* reg */
		if (width == 1) {
			if (modrm < 4) {
				v = r->reg32[modrm] & 0xff;
			} else {
				v = (r->reg32[modrm - 4] >> 8) & 0xff;
			}
		}
		if (width == 2) {
			v = r->reg32[modrm] & 0xffff;
		}
		if (width == 4) {
			v = r->reg32[modrm];
		}
	}
	if (mod < 3) {	/* mem */
		(*(r->panic))(r, "he86_read_regmem: mem");
	}
	return v;
}

void he86_loadsreg(struct HE86_REGS *r, int sreg_no, int sel, int flags)
/* flags͏̊gp([hɗÕ`FbNȂȂ)Aʏgpł0Ƃ */
{
	if ((r->cr[0] & HE86_CR0_PE) != 0) {
		/* GDTă[hAA=1ɂ̂Yꂸ */
		(*(r->panic))(r, "he86_loadsreg: 32bit");
	} else {
		r->sreg[sreg_no] = sel;
		r->segbase[sreg_no] = sel << 4;
	}
	return;
}

void he86_init_rm(struct HE86_REGS *r)
/* A[hpɐݒ */
{
	int i;
	for (i = 0; i < 6; i++) {
		he86_loadsreg(r, i, 0, 0);
		r->seglimit[i] = 0xffff;
		r->segflags[i] = 0;
	}
	for (i = 0; i < 8; i++) {
		r->reg32[i] = 0;
		r->cr[i] = 0;
		r->dr[i] = 0;
		r->tr[i] = 0;
	}
	r->eip = 0;
	r->eflags = 0x0202;
	r->step = 0;
	r->sysflags = HE86_SFLG_USE16C | HE86_SFLG_USE16S;
	return;
}
